==> milter-callback-1.5.0.tar.bz2: dealwithmxes.c <==
/* milter-callback - a sendmail milter to perform variuos sender verifications
 Copyright (C) 2008  Eugene M. Zheganin, milter-callback@norma.perm.ru

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/   

#include <milter-callback.h>
#include <commonprocs.h>

extern int logLevel;
extern int gracefulDNS;
extern char *logString;

int
dealWithMXes (char *emailaddress, char *hostPart, struct mlfiPriv *privData) {

/*
    Return Codes:
    0 - all of the MXes accepted emailAddress.
    1 - some of the MXes accepted emailAddress.
    2 - no MX for emailAddress found.
    3 - some software/memory error happened.
    4 - all MXes reject
    5 - some of the MXes tempfails
*/

    char emailBuf[MAXEMAIL];
    int i, dup, msgcount, returnCode;
    
    ns_msg handle;  /* handle for response packet */
    int rrnum;  /* resource record number */
    ns_rr rr;   /* expanded resource record */
    ns_sect section;
    
    #define MAX_MX 16
    
    /* MX array */
    char *mxList[MAX_MX];
    /* MX array length */
    int  mxNum = 0;
    
    union {
        HEADER hdr;              /* defined in resolv.h */
        u_char buf[NS_PACKETSZ]; /* defined in arpa/nameser.h */
    } response;                  /* response buffers */
    
    int responseLen;             /* buffer length */
    
    struct addrinfo hints, *res, *res0;
    int resError;
    int bufLen;
    
    char *relayVect;
    char *relayVectBeginsAt;
    char afterDog[MAXEMAIL];
    char *afterDogPointer;
	
    int relayRes2DogPart, domLevel;
    
    bzero(emailBuf, sizeof(emailBuf));
    
    snprintf(logString, MAXLOGLINE, "Entered dealWithMXes().");
    logNow(logString, 3);
    
    snprintf(logString, MAXLOGLINE, "Got an e-mail, %s.", emailaddress);
    logNow(logString, 3);
    snprintf(logString, MAXLOGLINE, "Got a host: '%s', %d symbols.", hostPart, strlen(hostPart));
    logNow(logString, 3);
    
    if ((responseLen = res_query(hostPart, ns_c_in, ns_t_mx, (u_char *) &response, sizeof(response))) < 0) {
	snprintf(logString, MAXLOGLINE, "Error while resolving %s, got \"%s\".", hostPart, hstrerror(h_errno));
	logNow(logString, 0);
	if (h_errno == 4) {
	    /* no MX found */
	    privData->MXesNum = 0;
	    privData->MXesThatAccept = 0;
	    privData->MXesThatReject = 0;
	    privData->MXesThatTempfail = 0;
	    privData->MXesOutOfOrder = 0;
	    snprintf(logString, MAXLOGLINE, "No MXes for the domain, %s.", hostPart);
    	    logNow(logString, 0);
	    return(2);
	}
    } else {
	snprintf(logString, MAXLOGLINE, "Successfully queried DNS.");
	logNow(logString, 3);
	snprintf(logString, MAXLOGLINE, "Response length is %d, or %d bytes.", responseLen, sizeof(response));
	logNow(logString, 3);
	snprintf(logString, MAXLOGLINE, "Size of response.buf was %d.", sizeof(response.buf));
	logNow(logString, 3);
	
	if (ns_initparse(response.buf, responseLen, &handle) < 0) {
	    snprintf(logString, MAXLOGLINE, "Error parsing response: %s.", strerror(errno));
	    logNow(logString, 0);
	    return(3);
	} else {	
	    section = ns_s_an;
	    msgcount = ns_msg_count(handle, section);
	    snprintf(logString, MAXLOGLINE, "Got %d records in answer section.", msgcount);
	    logNow(logString, 3);
	    for(rrnum = 0; rrnum < msgcount; rrnum++) {
		/*
	        * Expand the resource record number rrnum into rr.
		*/
		snprintf(logString, MAXLOGLINE, "Parsing record no %d.", rrnum);
		logNow(logString, 3);
		
		if (ns_parserr(&handle, section, rrnum, &rr) < 0) {
		    snprintf(logString, MAXLOGLINE, "ns_parserr: %s.", strerror(errno));
		    logNow(logString, 0);
		    return(3);
		} else {
		    /*
		    * If the record type is MX, save the name of the
		    * name server.
		    */
		    if (ns_rr_type(rr) == ns_t_mx) {
			snprintf(logString, MAXLOGLINE, "We got an MX record.");
			logNow(logString, 3);
			
			mxList[mxNum] = (char *) malloc(MAXDNAME);
			if(mxList[mxNum] == NULL){
			    snprintf(logString, MAXLOGLINE, "Cannot allocate memory.");
			    logNow(logString, 0);
			    return(3);
			}
			/* Expand the MX name */
			if (ns_name_uncompress(
			    ns_msg_base(handle),/* Start of the packet   */
			    ns_msg_end(handle), /* End of the packet     */
		    	    ns_rr_rdata(rr) + 2,    /* Position in the packet*/
			    mxList[mxNum],     /* Result                */
			    MAXDNAME)           /* Size of nsList buffer */
			    	    < 0) {    /* Negative: error       */
			    snprintf(logString, MAXLOGLINE, "ns_name_uncompress failed.");
			    logNow(logString, 0);
			    return(3);
			}
			/*
		        * Check the name we've just unpacked and add it to
		        * the list of MXes if it is not a duplicate.
		        * If it is a duplicate, just ignore it.
		        */
			for(i = 0, dup = 0; (i < mxNum) && !dup; i++) {
			    snprintf(logString, MAXLOGLINE, "Comparing %s against %s.", mxList[mxNum], mxList[i]);
			    logNow(logString, 3);
			    dup = !strncasecmp(mxList[i], mxList[mxNum], strlen(mxList[i]));
			}
			if(dup) {
			    snprintf(logString, MAXLOGLINE, "We got a duplicate.");
			    logNow(logString, 3);
			    if (mxList[mxNum] != NULL) {
				free(mxList[mxNum]);
			    }
			} else {
			    mxNum++;
			}
		    } else {
			snprintf(logString, MAXLOGLINE, "We got a RR of type %d, skipping in wait for %d.", ns_rr_type(rr), ns_t_mx);
			logNow(logString, 3);
		    }
		}
	    }
	    snprintf(logString, MAXLOGLINE, "Got %d MXes:", mxNum);
	    logNow(logString, 2);
	    for (i = 0; i < mxNum; i++) {
		snprintf(logString, MAXLOGLINE, "%d: %s", i, mxList[i]);
		logNow(logString, 2);
	    }
	}
    }
    /* here we already got the list of MXes */
    snprintf(logString, MAXLOGLINE, "Commencing the mxlist parsing.");
    logNow(logString, 2);
    
    /* determining if relay is one on the MXes */
    for(i = 0; i < mxNum && privData->skipDirect == 0; i++) {
	snprintf(logString, MAXLOGLINE, "Checking MX %d of %d, %s against the relay IP, %s.", i + 1, mxNum, mxList[i], privData->relayPresIP);
	logNow(logString, 2);
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	resError = getaddrinfo(mxList[i], NULL, &hints, &res0);
	if (resError) {
	    snprintf(logString, MAXLOGLINE, "Cannot resolve host %s: %s", mxList[i], gai_strerror(resError));
	    logNow(logString, 2);
	} else {
	    for (res = res0; res; res = res->ai_next) {
		if (((struct sockaddr_in *)res->ai_addr)->sin_addr.s_addr == privData->relayIP) {
		    snprintf(logString, MAXLOGLINE, "Setting skipDirect flag for relay %s - it's IP is identical to MX IP %s", privData->relayPresIP, mxList[i]);
		    logNow(logString, 2);
		    privData->skipDirect = 1;
		}
	    }
	    freeaddrinfo(res0);
	}
    }
    
    if (privData->skipDirect == 0 && gracefulDNS == 1) {
	/* determinig if the relay IP resolves to the name containing full hostpart */
	relayRes2DogPart = 0 ;
	domLevel = 1;
	/* pushing email and relay*/
	bufLen = strlen(privData->resolvedRelayIP) + 1;
	relayVect = malloc(bufLen);
	bzero(relayVect, bufLen);
	strncpy(relayVect, privData->resolvedRelayIP, bufLen);	
	bzero(afterDog, sizeof(afterDog));
	strncpy(afterDog, emailaddress, sizeof(afterDog));
	if (relayVect == NULL) {
	    snprintf(logString, MAXLOGLINE, "Cannot allocate memory.");
	    logNow(logString, 0);
	} else {
	    relayVectBeginsAt = relayVect;
	    afterDogPointer = &afterDog[0];
	    strsep(&afterDogPointer, "@");
	    snprintf(logString, MAXLOGLINE, "afterDogPointer is %s.", afterDogPointer);
	    logNow(logString, 3);
	    while (relayVect != NULL && relayRes2DogPart == 0) {
		snprintf(logString, MAXLOGLINE, "relayVect is %s.", relayVect);
		logNow(logString, 3);
    		if (relayVect != NULL && strcmp(relayVect, afterDogPointer) == 0) {
		    snprintf(logString, MAXLOGLINE, "%s is equal to %s.", relayVect, afterDogPointer);
		    logNow(logString, 3);
        	    relayRes2DogPart = 1;
    		}
		strsep(&relayVect, ".");
	    }
	    if (relayRes2DogPart == 1) {
		/* finding the level of the domain */
		for (i = 0; i < strlen(afterDogPointer); i++) {
    		    if (afterDogPointer[i] == '.') {
    			domLevel++;
    	    	    }
		}
		snprintf(logString, MAXLOGLINE, "Domain level of e-mail hostpart %s is %d.", afterDogPointer, domLevel);
		logNow(logString, 3);
		if (domLevel >= 2) {
		    snprintf(logString, MAXLOGLINE, "Setting graceful DNS bit, [%s] resolves to %s, which contains hostpart %s.", privData->relayPresIP, privData->resolvedRelayIP, afterDogPointer);
		    logNow(logString, 3);
		    privData->gracefulDNS = 1;
		}
	    }
	    if (relayVectBeginsAt != NULL) {
		snprintf(logString, MAXLOGLINE, "Trying to free relayVectBeginsAt in dealWithMXes().");
		logNow(logString, 3);
		free(relayVectBeginsAt);
	    }
	}
    } else {
	if (privData->skipDirect == 1) {
	    snprintf(logString, MAXLOGLINE, "Skipped graceful DNS check, skipDirect flag already set.");
	    logNow(logString, 3);
	}
    }
    
    privData->MXesNum = mxNum;
    privData->MXesThatAccept = 0;
    privData->MXesThatReject = 0;
    privData->MXesThatTempfail = 0;
    privData->MXesOutOfOrder = 0;
    
    for(i = 0; i < mxNum; i++) {
	snprintf(logString, MAXLOGLINE, "Parsing relay %d of %d, %s.", i + 1, mxNum, mxList[i]);
	logNow(logString, 2);
	returnCode = checkServerResponse(emailaddress, mxList[i], NULL, 0);
	snprintf(logString, MAXLOGLINE, "checkServerResponse() returned %d.", returnCode);
	logNow(logString, 3);
	switch (returnCode) {
	    case 0:
		/* 0 - email accepped */
		snprintf(logString, MAXLOGLINE, "Got one of the accepting MXes.");
		logNow(logString, 2);
		privData->MXesThatAccept++;
		break;
	    case 1:
		/* 1 - email not accepted */
		snprintf(logString, MAXLOGLINE, "Got one of the rejecting MXes.");
		logNow(logString, 2);
		privData->MXesThatReject++;
		break;
	    case 2:
		/* 2 - email tempfailed */
		snprintf(logString, MAXLOGLINE, "Got one of the tempfailing MXes.");
		logNow(logString, 2);
		privData->MXesThatTempfail++;
		break;
	    case 3:
		/* 3 - socket error */
		snprintf(logString, MAXLOGLINE, "Got a socket error.");
		logNow(logString, 2);
		privData->MXesOutOfOrder++;
		break;
	    case 4:
		/* 4 - DNS error */
		snprintf(logString, MAXLOGLINE, "Got a DNS error.");
		privData->MXesOutOfOrder++;
		logNow(logString, 2);
		break;
	    case 5:
		/* 5 - unknown answer code */
		snprintf(logString, MAXLOGLINE, "Got an unknown SMTP answer code.");
		logNow(logString, 0);
		privData->MXesOutOfOrder++;
		break;
	    case 6:
	    default:
		/* 6 - memory error */
		snprintf(logString, MAXLOGLINE, "Got a memory error.");
		logNow(logString, 0);
		privData->MXesOutOfOrder++;
		break;
	}
    }
    /* freeing mxList */
    snprintf(logString, MAXLOGLINE, "Trying to free mxList in dealWithMXes().");
    logNow(logString, 3);
    for (i = 0; i < mxNum; i++) {
	if (mxList[i] != NULL) {
	    snprintf(logString, MAXLOGLINE, "freeing MX %d in dealWithMXes().", i);
	    logNow(logString, 3);
	    free(mxList[i]);
	}
    }
    
    if (privData->MXesThatAccept == mxNum) {
	return(0);
    } else {
	if (privData->MXesThatAccept > 0){
	    return(1);
	} else {
	    if (privData->MXesThatTempfail > 0) {
		return(5);
	    } else {
		if (privData->MXesOutOfOrder > 0) {
		    return(3);
		} else {
		    return(4);
		}
	    }
	}
    }
}

==> milter-callback-1.5.2.tar.bz2: dealwithmxes.c <==
/* milter-callback - a sendmail milter to perform variuos sender verifications
 Copyright (C) 2008  Eugene M. Zheganin, milter-callback@norma.perm.ru

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/   

#include <milter-callback.h>
#include <commonprocs.h>

extern int logLevel;
extern int gracefulDNS;
extern char *logString;

int
dealWithMXes (char *emailaddress, char *hostPart, struct mlfiPriv *privData) {

/*
    Return Codes:
    0 - all of the MXes accepted emailAddress.
    1 - some of the MXes accepted emailAddress.
    2 - no MX for emailAddress found.
    3 - some software/memory error happened.
    4 - all MXes reject
    5 - some of the MXes tempfails
*/

    char emailBuf[MAXEMAIL];
    int i, dup, msgcount, returnCode;
    
    ns_msg handle;  /* handle for response packet */
    int rrnum;  /* resource record number */
    ns_rr rr;   /* expanded resource record */
    ns_sect section;
    
    #define MAX_MX 16
    
    /* MX array */
    char *mxList[MAX_MX];
    /* MX array length */
    int  mxNum = 0;
    
    union {
        HEADER hdr;              /* defined in resolv.h */
        u_char buf[NS_PACKETSZ]; /* defined in arpa/nameser.h */
    } response;                  /* response buffers */
    
    int responseLen;             /* buffer length */
    
    struct addrinfo hints, *res, *res0;
    int resError;
    int bufLen;
    
    char *relayVect;
    char *relayVectBeginsAt;
    char afterDog[MAXEMAIL];
    char *afterDogPointer;
	
    int relayRes2DogPart, domLevel;
    
    bzero(emailBuf, sizeof(emailBuf));
    
    snprintf(logString, MAXLOGLINE, "Entered dealWithMXes().");
    logNow(logString, 3, privData -> threadCounter);
    
    snprintf(logString, MAXLOGLINE, "Got an e-mail, %s.", emailaddress);
    logNow(logString, 3, privData -> threadCounter);
    snprintf(logString, MAXLOGLINE, "Got a host: '%s', %d symbols.", hostPart, strlen(hostPart));
    logNow(logString, 3, privData -> threadCounter);
    
    if ((responseLen = res_query(hostPart, ns_c_in, ns_t_mx, (u_char *) &response, sizeof(response))) < 0) {
	snprintf(logString, MAXLOGLINE, "Error while resolving %s, got \"%s\".", hostPart, hstrerror(h_errno));
	logNow(logString, 0, privData -> threadCounter);
	if (h_errno == 4) {
	    /* no MX found */
	    privData->MXesNum = 0;
	    privData->MXesThatAccept = 0;
	    privData->MXesThatReject = 0;
	    privData->MXesThatTempfail = 0;
	    privData->MXesOutOfOrder = 0;
	    snprintf(logString, MAXLOGLINE, "No MXes for the domain, %s.", hostPart);
    	    logNow(logString, 0, privData -> threadCounter);
	    return(2);
	}
    } else {
	snprintf(logString, MAXLOGLINE, "Successfully queried DNS.");
	logNow(logString, 3, privData -> threadCounter);
	snprintf(logString, MAXLOGLINE, "Response length is %d, or %d bytes.", responseLen, sizeof(response));
	logNow(logString, 3, privData -> threadCounter);
	snprintf(logString, MAXLOGLINE, "Size of response.buf was %d.", sizeof(response.buf));
	logNow(logString, 3, privData -> threadCounter);
	
	if (ns_initparse(response.buf, responseLen, &handle) < 0) {
	    snprintf(logString, MAXLOGLINE, "Error parsing response: %s.", strerror(errno));
	    logNow(logString, 0, privData -> threadCounter);
	    return(3);
	} else {	
	    section = ns_s_an;
	    msgcount = ns_msg_count(handle, section);
	    snprintf(logString, MAXLOGLINE, "Got %d records in answer section.", msgcount);
	    logNow(logString, 3, privData -> threadCounter);
	    for(rrnum = 0; rrnum < msgcount; rrnum++) {
		/*
	        * Expand the resource record number rrnum into rr.
		*/
		snprintf(logString, MAXLOGLINE, "Parsing record no %d.", rrnum);
		logNow(logString, 3, privData -> threadCounter);
		
		if (ns_parserr(&handle, section, rrnum, &rr) < 0) {
		    snprintf(logString, MAXLOGLINE, "ns_parserr: %s.", strerror(errno));
		    logNow(logString, 0, privData -> threadCounter);
		    return(3);
		} else {
		    /*
		    * If the record type is MX, save the name of the
		    * name server.
		    */
		    if (ns_rr_type(rr) == ns_t_mx) {
			snprintf(logString, MAXLOGLINE, "We got an MX record.");
			logNow(logString, 3, privData -> threadCounter);
			
			mxList[mxNum] = (char *) malloc(MAXDNAME);
			if(mxList[mxNum] == NULL){
			    snprintf(logString, MAXLOGLINE, "Cannot allocate memory.");
			    logNow(logString, 0, privData -> threadCounter);
			    return(3);
			}
			/* Expand the MX name */
			if (ns_name_uncompress(
			    ns_msg_base(handle),/* Start of the packet   */
			    ns_msg_end(handle), /* End of the packet     */
		    	    ns_rr_rdata(rr) + 2,    /* Position in the packet*/
			    mxList[mxNum],     /* Result                */
			    MAXDNAME)           /* Size of nsList buffer */
			    	    < 0) {    /* Negative: error       */
			    snprintf(logString, MAXLOGLINE, "ns_name_uncompress failed.");
			    logNow(logString, 0, privData -> threadCounter);
			    return(3);
			}
			/*
		        * Check the name we've just unpacked and add it to
		        * the list of MXes if it is not a duplicate.
		        * If it is a duplicate, just ignore it.
		        */
			for(i = 0, dup = 0; (i < mxNum) && !dup; i++) {
			    snprintf(logString, MAXLOGLINE, "Comparing %s against %s.", mxList[mxNum], mxList[i]);
			    logNow(logString, 3, privData -> threadCounter);
			    dup = !strncasecmp(mxList[i], mxList[mxNum], strlen(mxList[i]));
			}
			if(dup) {
			    snprintf(logString, MAXLOGLINE, "We got a duplicate.");
			    logNow(logString, 3, privData -> threadCounter);
			    if (mxList[mxNum] != NULL) {
				free(mxList[mxNum]);
			    }
			} else {
			    mxNum++;
			}
		    } else {
			snprintf(logString, MAXLOGLINE, "We got a RR of type %d, skipping in wait for %d.", ns_rr_type(rr), ns_t_mx);
			logNow(logString, 3, privData -> threadCounter);
		    }
		}
	    }
	    snprintf(logString, MAXLOGLINE, "Got %d MXes:", mxNum);
	    logNow(logString, 2, privData -> threadCounter);
	    for (i = 0; i < mxNum; i++) {
		snprintf(logString, MAXLOGLINE, "%d: %s", i, mxList[i]);
		logNow(logString, 2, privData -> threadCounter);
	    }
	}
    }
    /* here we already got the list of MXes */
    snprintf(logString, MAXLOGLINE, "Commencing the mxlist parsing.");
    logNow(logString, 2, privData -> threadCounter);
    
    /* determining if relay is one on the MXes */
    for(i = 0; i < mxNum && privData->skipDirect == 0; i++) {
	snprintf(logString, MAXLOGLINE, "Checking MX %d of %d, %s against the relay IP, %s.", i + 1, mxNum, mxList[i], privData->relayPresIP);
	logNow(logString, 2, privData -> threadCounter);
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	resError = getaddrinfo(mxList[i], NULL, &hints, &res0);
	if (resError) {
	    snprintf(logString, MAXLOGLINE, "Cannot resolve host %s: %s", mxList[i], gai_strerror(resError));
	    logNow(logString, 2, privData -> threadCounter);
	} else {
	    for (res = res0; res; res = res->ai_next) {
		if (((struct sockaddr_in *)res->ai_addr)->sin_addr.s_addr == privData->relayIP) {
		    snprintf(logString, MAXLOGLINE, "Setting skipDirect flag for relay %s - it's IP is identical to MX IP %s", privData->relayPresIP, mxList[i]);
		    logNow(logString, 2, privData -> threadCounter);
		    privData->skipDirect = 1;
		}
	    }
	    freeaddrinfo(res0);
	}
    }
    
    if (privData->skipDirect == 0 && gracefulDNS == 1) {
	/* determinig if the relay IP resolves to the name containing full hostpart */
	relayRes2DogPart = 0 ;
	domLevel = 1;
	/* pushing email and relay*/
	bufLen = strlen(privData->resolvedRelayIP) + 1;
	relayVect = malloc(bufLen);
	bzero(relayVect, bufLen);
	strncpy(relayVect, privData->resolvedRelayIP, bufLen);	
	bzero(afterDog, sizeof(afterDog));
	strncpy(afterDog, emailaddress, sizeof(afterDog));
	if (relayVect == NULL) {
	    snprintf(logString, MAXLOGLINE, "Cannot allocate memory.");
	    logNow(logString, 0, privData -> threadCounter);
	} else {
	    relayVectBeginsAt = relayVect;
	    afterDogPointer = &afterDog[0];
	    strsep(&afterDogPointer, "@");
	    snprintf(logString, MAXLOGLINE, "afterDogPointer is %s.", afterDogPointer);
	    logNow(logString, 3, privData -> threadCounter);
	    while (relayVect != NULL && relayRes2DogPart == 0) {
		snprintf(logString, MAXLOGLINE, "relayVect is %s.", relayVect);
		logNow(logString, 3, privData -> threadCounter);
    		if (relayVect != NULL && strcmp(relayVect, afterDogPointer) == 0) {
		    snprintf(logString, MAXLOGLINE, "%s is equal to %s.", relayVect, afterDogPointer);
		    logNow(logString, 3, privData -> threadCounter);
        	    relayRes2DogPart = 1;
    		}
		strsep(&relayVect, ".");
	    }
	    if (relayRes2DogPart == 1) {
		/* finding the level of the domain */
		for (i = 0; i < strlen(afterDogPointer); i++) {
    		    if (afterDogPointer[i] == '.') {
    			domLevel++;
    	    	    }
		}
		snprintf(logString, MAXLOGLINE, "Domain level of e-mail hostpart %s is %d.", afterDogPointer, domLevel);
		logNow(logString, 3, privData -> threadCounter);
		if (domLevel >= 2) {
		    snprintf(logString, MAXLOGLINE, "Setting graceful DNS bit, [%s] resolves to %s, which contains hostpart %s.", privData->relayPresIP, privData->resolvedRelayIP, afterDogPointer);
		    logNow(logString, 3, privData -> threadCounter);
		    privData->gracefulDNS = 1;
		}
	    }
	    if (relayVectBeginsAt != NULL) {
		snprintf(logString, MAXLOGLINE, "Trying to free relayVectBeginsAt in dealWithMXes().");
		logNow(logString, 3, privData -> threadCounter);
		free(relayVectBeginsAt);
	    }
	}
    } else {
	if (privData->skipDirect == 1) {
	    snprintf(logString, MAXLOGLINE, "Skipped graceful DNS check, skipDirect flag already set.");
	    logNow(logString, 3, privData -> threadCounter);
	}
    }
    
    privData->MXesNum = mxNum;
    privData->MXesThatAccept = 0;
    privData->MXesThatReject = 0;
    privData->MXesThatTempfail = 0;
    privData->MXesOutOfOrder = 0;
    
    for(i = 0; i < mxNum; i++) {
	snprintf(logString, MAXLOGLINE, "Parsing relay %d of %d, %s.", i + 1, mxNum, mxList[i]);
	logNow(logString, 2, privData -> threadCounter);
	returnCode = checkServerResponse(emailaddress, mxList[i], NULL, 0, privData);
	snprintf(logString, MAXLOGLINE, "checkServerResponse() returned %d.", returnCode);
	logNow(logString, 3, privData -> threadCounter);
	switch (returnCode) {
	    case 0:
		/* 0 - email accepped */
		snprintf(logString, MAXLOGLINE, "Got one of the accepting MXes.");
		logNow(logString, 2, privData -> threadCounter);
		privData->MXesThatAccept++;
		break;
	    case 1:
		/* 1 - email not accepted */
		snprintf(logString, MAXLOGLINE, "Got one of the rejecting MXes.");
		logNow(logString, 2, privData -> threadCounter);
		privData->MXesThatReject++;
		break;
	    case 2:
		/* 2 - email tempfailed */
		snprintf(logString, MAXLOGLINE, "Got one of the tempfailing MXes.");
		logNow(logString, 2, privData -> threadCounter);
		privData->MXesThatTempfail++;
		break;
	    case 3:
		/* 3 - socket error */
		snprintf(logString, MAXLOGLINE, "Got a socket error.");
		logNow(logString, 2, privData -> threadCounter);
		privData->MXesOutOfOrder++;
		break;
	    case 4:
		/* 4 - DNS error */
		snprintf(logString, MAXLOGLINE, "Got a DNS error.");
		privData->MXesOutOfOrder++;
		logNow(logString, 2, privData -> threadCounter);
		break;
	    case 5:
		/* 5 - unknown answer code */
		snprintf(logString, MAXLOGLINE, "Got an unknown SMTP answer code.");
		logNow(logString, 0, privData -> threadCounter);
		privData->MXesOutOfOrder++;
		break;
	    case 6:
	    default:
		/* 6 - memory error */
		snprintf(logString, MAXLOGLINE, "Got a memory error.");
		logNow(logString, 0, privData -> threadCounter);
		privData->MXesOutOfOrder++;
		break;
	}
    }
    /* freeing mxList */
    snprintf(logString, MAXLOGLINE, "Trying to free mxList in dealWithMXes().");
    logNow(logString, 3, privData -> threadCounter);
    for (i = 0; i < mxNum; i++) {
	if (mxList[i] != NULL) {
	    snprintf(logString, MAXLOGLINE, "freeing MX %d in dealWithMXes().", i);
	    logNow(logString, 3, privData -> threadCounter);
	    free(mxList[i]);
	}
    }
    
    if (privData->MXesThatAccept == mxNum) {
	return(0);
    } else {
	if (privData->MXesThatAccept > 0){
	    return(1);
	} else {
	    if (privData->MXesThatTempfail > 0) {
		return(5);
	    } else {
		if (privData->MXesOutOfOrder > 0) {
		    return(3);
		} else {
		    return(4);
		}
	    }
	}
    }
}
